/*****************************************************************************
 *
 * Copyright (C) 2014      Advanced Digital Chips, Inc. All Rights Reserved.
 *						http://www.adc.co.kr
 *
 * THIS SOFTWARE IS PROVIDED BY ADCHIPS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ADCHIPS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE
 *
 *****************************************************************************/

#include "sdk.h"

#define FLASH_WRITE_ENABLE 1

#if FLASH_WRITE_ENABLE
#define WRITE_PROTECT_ENABLE(N) \
	do                          \
	{                           \
		*R_FLCMD(N) = 0x6;      \
		*R_FLSTS(N) = 0x3C;     \
		sflash_wait_busy(N);    \
	} while (0)
#define WRITE_PROTECT_DISABLE(N) \
	do                           \
	{                            \
		*R_FLCMD(N) = 0x6;       \
		*R_FLSTS(N) = 0x00;      \
		sflash_wait_busy(N);     \
	} while (0)
#else
#define WRITE_PROTECT_ENABLE(N)
#define WRITE_PROTECT_DISABLE(N)
#endif

typedef struct _flashdbtag
{
	BYTE id1;
	BYTE id2;
	BYTE id3;
	DWORD sectorsize;
	DWORD blocksize;
	DWORD memsize;
	BOOL bQuad;
	char strdesc[128];
} FLASHDB;
static const FLASHDB predef_db[] = {
	//  EON
	{ 0x1c, 0x31, 0x12, 4 * 1024, 64 * 1024, 256 * 1024, FALSE, "EON" },
	{ 0x1c, 0x30, 0x13, 4 * 1024, 64 * 1024, 512 * 1024, TRUE, "EON" },
	{ 0x1c, 0x31, 0x13, 4 * 1024, 64 * 1024, 512 * 1024, FALSE, "EON" },
	{ 0x1c, 0x30, 0x14, 4 * 1024, 64 * 1024, 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x31, 0x14, 4 * 1024, 64 * 1024, 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x30, 0x15, 4 * 1024, 64 * 1024, 2 * 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x30, 0x16, 4 * 1024, 64 * 1024, 4 * 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x30, 0x17, 4 * 1024, 64 * 1024, 8 * 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x30, 0x18, 4 * 1024, 64 * 1024, 16 * 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x70, 0x15, 4 * 1024, 64 * 1024, 2 * 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x70, 0x16, 4 * 1024, 64 * 1024, 4 * 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x70, 0x17, 4 * 1024, 64 * 1024, 8 * 1024 * 1024, TRUE, "EON" },
	{ 0x1c, 0x70, 0x18, 4 * 1024, 64 * 1024, 16 * 1024 * 1024, TRUE, "EON" },

	//  WINBOND
	{ 0xef, 0x30, 0x11, 4 * 1024, 64 * 1024, 512 * 1024, FALSE, "WINBOND" },
	{ 0xef, 0x30, 0x12, 4 * 1024, 64 * 1024, 512 * 1024, FALSE, "WINBOND" },
	{ 0xef, 0x30, 0x13, 4 * 1024, 64 * 1024, 512 * 1024, FALSE, "WINBOND" },
	{ 0xef, 0x30, 0x14, 4 * 1024, 64 * 1024, 1024 * 1024, FALSE, "WINBOND" },
	{ 0xef, 0x30, 0x15, 4 * 1024, 64 * 1024, 2 * 1024 * 1024, FALSE,
	  "WINBOND" },
	{ 0xef, 0x30, 0x16, 4 * 1024, 64 * 1024, 4 * 1024 * 1024, FALSE,
	  "WINBOND" },

	{ 0xef, 0x40, 0x12, 4 * 1024, 64 * 1024, 256 * 1024, TRUE, "WINBOND" },
	{ 0xef, 0x40, 0x13, 4 * 1024, 64 * 1024, 512 * 1024, TRUE, "WINBOND" },
	{ 0xef, 0x40, 0x14, 4 * 1024, 64 * 1024, 1024 * 1024, TRUE, "WINBOND" },
	{ 0xef, 0x40, 0x15, 4 * 1024, 64 * 1024, 2 * 1024 * 1024, TRUE, "WINBOND" },
	{ 0xef, 0x40, 0x16, 4 * 1024, 64 * 1024, 4 * 1024 * 1024, TRUE, "WINBOND" },
	{ 0xef, 0x40, 0x17, 4 * 1024, 32 * 1024, 8 * 1024 * 1024, FALSE,
	  "WINBOND" },
	{ 0xef, 0x40, 0x18, 4 * 1024, 64 * 1024, 16 * 1024 * 1024, TRUE,
	  "WINBOND" },

	{ 0, 0, 0, 0, 0, 0 }, //  MAX
};

static struct _tagFLASH
{
	U8 ids[4];
	U32 sector_size;
	U32 block_size;
	U32 size;
	BOOL bQuad;
	U32 xip_addr;
} flash_info[2] = {
	{
		{
			0,
		},
	},
};

void RAMCODE sflash_wait_busy(int dev_num)
{
	while (1)
	{
		vU8 stat = *R_FLSTS(dev_num);
		if ((stat & 1) == 0)
			break;
	}
}

void RAMCODE sflash_write_status(int dev_num, U8 cmd, U8 data)
{
	CRITICAL_ENTER();
	*R_FLCMD(dev_num)		 = 0x6; // write enable
	U32 org_mod				 = *R_FLMOD(dev_num);
	*R_FLMOD(dev_num)		 = 0; // CS LOW, write-enable
	*(vU8 *)R_FLDAT(dev_num) = cmd;
	*(vU8 *)R_FLDAT(dev_num) = data;
	*R_FLMOD(dev_num)		 = org_mod;
	sflash_wait_busy(dev_num);
	CRITICAL_EXIT();
}

RAMCODE void sflash_erase_sector(int dev_num, int sector, int sector_cnt)
{
	int i;
	U32 mode		  = *R_FLMOD(dev_num);
	*R_FLMOD(dev_num) = mode & (~(1 << 7)); // write enable
	WRITE_PROTECT_DISABLE(dev_num);

	U32 addr = sector * flash_info[dev_num].sector_size;
	for (i = 0; i < sector_cnt; i++)
	{
		sflash_custom_write(dev_num, 0x20, addr, 0, 0);
		addr += flash_info[dev_num].sector_size;
	}

	WRITE_PROTECT_ENABLE(dev_num);
	*R_FLMOD(dev_num) = mode;
}
RAMCODE void sflash_dev_write(U32 dev_num, U32 offset, BYTE *buf, int len)
{
	WRITE_PROTECT_DISABLE(dev_num);

	if (offset & 0xFF) // 256byte miss align
	{
		int write_len = 256 - (offset & 0xff);
		if (write_len > len)
			write_len = len;

		sflash_custom_write(dev_num, 0x2, offset, buf, write_len);
		len -= write_len;
		buf += write_len;
		offset += write_len;
	}

	while (1)
	{
		if (len > 256)
		{
			sflash_custom_write(dev_num, 0x2, offset, buf, 256);
		}
		else
		{
			sflash_custom_write(dev_num, 0x2, offset, buf, len);
			break;
		}

		len -= 256;
		buf += 256;
		offset += 256;
	}
	icache_invalidate_way();
	dcache_invalidate_way();
	WRITE_PROTECT_ENABLE(dev_num);
}
RAMCODE void sflash_write(U32 addr, BYTE *buf, int len)
{
	int dev_num = 0;
	if ((addr >= flash_info[0].size) && (addr < flash_info[1].xip_addr))
	{
		debugstring("invalid address of flash memory\r\n");
		return;
	}
	if (addr >= flash_info[1].xip_addr)
	{
		addr -= flash_info[1].xip_addr;
		dev_num = 1;
	}
	sflash_dev_write(dev_num, addr, buf, len);
}

int RAMCODE sflash_read(U32 addr, BYTE *buf, int len)
{
	int dev_num = 0;
	if ((addr >= flash_info[0].size) && (addr < flash_info[1].xip_addr))
	{
		debugstring("invalid address of flash memory\r\n");
		return 0;
	}
	if (addr >= flash_info[1].xip_addr)
	{
		addr -= flash_info[1].xip_addr;
		dev_num = 1;
	}

	sflash_custom_read(dev_num, 0x3, addr, 3, 0, buf, len);

	return len;
}
int RAMCODE sflash_dev_read(int dev_num, U32 offset, BYTE *buf, int len)
{
	return sflash_custom_read(dev_num, 0x3, offset, 3, 0, buf, len);
}

int sflash_get_size(int dev_num)
{
	return flash_info[dev_num].size;
}
int sflash_get_sector_size(int dev_num)
{
	return flash_info[dev_num].sector_size;
}
int sflash_get_sector_count(int dev_num)
{
	return flash_info[dev_num].size/flash_info[dev_num].sector_size;
}
RAMCODE static void read_id(int dev_num)
{
	CRITICAL_ENTER();
	U32 org;
	org						   = *R_FLMOD(dev_num);
	*R_FLMOD(dev_num)		   = 0;	   // CS LOW FIX, write enable
	*(vU8 *)R_FLDAT(dev_num)   = 0x9f; // jedec id
	flash_info[dev_num].ids[0] = *(vU8 *)R_FLDAT(dev_num);
	flash_info[dev_num].ids[1] = *(vU8 *)R_FLDAT(dev_num);
	flash_info[dev_num].ids[2] = *(vU8 *)R_FLDAT(dev_num);
	*R_FLMOD(dev_num)		   = org;
	CRITICAL_EXIT();
}

#ifdef SF_SUPPORTS_SFDP
// RAM Usage : 750 byte
// CODE Usage : 750 Bye
typedef struct _tagSFDP
{
	U8 signature[4];
	U8 sfdp_minor_ver;
	U8 sfdp_major_ver;
	U8 nph;
	U8 resv;
	// 1st parameter header
	U8 jedec_id;
	U8 param_minor_ver;
	U8 param_major_ver;
	U8 param_length;
	U32 ptp; // 3byte valid, Parameter Table Pointer(PTP)
} SFDP_HEADER;
SFDP_HEADER sfdp;
static U8 sfdp_param_table[32 * 4];

RAMCODE BOOL read_sfdp(int dev_num)
{
	CRITICAL_ENTER();
	BOOL re = TRUE;
	U32 org;
	org						 = *R_FLMOD(dev_num);
	*R_FLMOD(dev_num)		 = 0;	 // CS LOW FIX, write enable
	*(vU8 *)R_FLDAT(dev_num) = 0x5a; // SFDP
	*(vU8 *)R_FLDAT(dev_num) = 0x0;	 // address
	*(vU8 *)R_FLDAT(dev_num) = 0x0;	 // address
	*(vU8 *)R_FLDAT(dev_num) = 0x0;	 // address
	*(vU8 *)R_FLDAT(dev_num) = 0x0;	 // dummy

	int i;

	BYTE *pbuf = &sfdp.signature[0];
	for (i = 0; i < 16; i++)
		pbuf[i] = *(vU8 *)R_FLDAT(dev_num);

	if ((sfdp.signature[0] == 'S') && (sfdp.signature[1] == 'F') &&
		(sfdp.signature[2] == 'D') && (sfdp.signature[3] == 'P'))
	{
		// dummy read
		U8 dummy;
		sfdp.ptp &= 0xffffff;
		int skipsize = sfdp.ptp - sizeof(SFDP_HEADER);
		for (i = 0; i < skipsize; i++)
			dummy = *(vU8 *)R_FLDAT(dev_num);

		for (i = 0; i < sfdp.param_length * 4; i++)
			sfdp_param_table[i] = *(vU8 *)R_FLDAT(dev_num);

		U32 flash_size_in_bits	 = (*(U32 *)&sfdp_param_table[4]) + 1;
		flash_info[dev_num].size = flash_size_in_bits / 8;
		if ((sfdp_param_table[0] & 0x3) == 1)
		{
			flash_info[dev_num].sector_size = 4 * 1024;
		}
		else
		{
			re = false;
		}

		if (sfdp_param_table[1] != 0x20)
		{
			re = false;
		}
		flash_info[dev_num].block_size = 0;
		for (i = 0; i < 4; i++)
		{
			U8 size	  = sfdp_param_table[4 * 7 + i * 2];
			U8 opcode = sfdp_param_table[4 * 7 + i * 2 + 1];
			if ((size != 0) && (opcode == 0xd8))
			{
				DWORD blocksize = 1 << size;
				if (blocksize > flash_info[dev_num].block_size)
				{
					flash_info[dev_num].block_size = blocksize;
					break;
				}
			}
		}
	}
	else
	{
		re = false;
	}
	*R_FLMOD(dev_num) = org;
	CRITICAL_EXIT();
	return re;
}
#endif
static BOOL find_sflash_info(int dev_num)
{
	int i;
	for (i = 0; 1; i++)
	{
		if (predef_db[i].id1 == 0) // reach end ?
		{
			// just decode capacity and try SFDP
			flash_info[dev_num].sector_size = 4 * 1024;
			flash_info[dev_num].size = 1 << (flash_info[dev_num].ids[2] + 8);
			return FALSE;
		}

		if ((predef_db[i].id1 == flash_info[dev_num].ids[0]) &&
			(predef_db[i].id2 == flash_info[dev_num].ids[1]) &&
			(predef_db[i].id3 == flash_info[dev_num].ids[2]))
		{
			flash_info[dev_num].sector_size = predef_db[i].sectorsize;
			flash_info[dev_num].size		= predef_db[i].memsize;
			flash_info[dev_num].bQuad		= predef_db[i].bQuad;
			break;
		}
	}
	return TRUE;
}

RAMCODE void sflash_init(int dev_num)
{
	int i;
	read_id(dev_num);

	if (find_sflash_info(dev_num) == FALSE)
	{
#ifdef SF_SUPPORTS_SFDP
		read_sfdp(dev_num);
#endif
	}
	*R_FLMOD(dev_num) = 0x180; // 7bit:write disable
	// stat = *(vU8*)R_FLSTS2(dev_num);

	WRITE_PROTECT_ENABLE(dev_num);

	CRITICAL_ENTER();
	*R_FLMOD(dev_num) = 0x382; // 7bit:write disable, Quad read
	*R_FLBRT(dev_num) = 0x00;  // flash baud-rate
	*R_FLCSH(dev_num) = 0xf;
	CRITICAL_EXIT();
	for (i = 0; i < 10; i++)
	{
		if (flash_info[dev_num].size == ((32 * 1024) << i))
			break;
	}
	*R_FLMSIZE(dev_num) = i;
	
	if (dev_num == 1)
	{
		/*
		set largest memory size of you have. (internal or external)
		1Mbyte: 2
		2Mbyte: 3
		4Mbyte: 4
		8Mbyte: 5
		16Mbyte: 6
		*/
		
		if (*R_FLMSIZE(1) > *R_FLMSIZE(0))
		{
			for (i = 0; i < 7; i++)
			{
				if (flash_info[1].size == ((256 * 1024) << i))
					break;
			}
			*(U32*)0x80023c78 = i; //external flash start address

			flash_info[1].xip_addr = flash_info[1].size;
		}
		else
		{
			for (i = 0; i < 7; i++)
			{
				if (flash_info[0].size == ((256 * 1024) << i))
					break;
			}
			*(U32*)0x80023c78 = i; //external flash start address
			flash_info[1].xip_addr = flash_info[0].size;
		}
	}
	else
		flash_info[1].xip_addr = flash_info[0].size;
	/*
    CASE: 4+4Mbyte Memory
        second memory's start address for XIP : 0x400000
    case: 4+8Mbyte Memory
        second memory's start address for XIP : 0x800000
    case: 4+2Mbyte Memory
        second memory's start address for XIP : 0x400000
    */
}

void RAMCODE sflash_custom_write(int dev_num, U8 cmd, U32 address, U8 *buf,
								 U32 buflen)
{
	CRITICAL_ENTER();	
	*R_FLCMD(dev_num)		 = 0x6; // write enable
	U32 org_mod				 = *R_FLMOD(dev_num);
	*R_FLMOD(dev_num)		 = 0; // CS LOW, write-enable
	*(vU8 *)R_FLDAT(dev_num) = cmd;
	*(vU8 *)R_FLDAT(dev_num) = (address >> 16) & 0xff;
	*(vU8 *)R_FLDAT(dev_num) = (address >> 8) & 0xff;
	*(vU8 *)R_FLDAT(dev_num) = address & 0xff;
	U32 i;
	for (i = 0; i < buflen; i++)
	{
		*(vU8 *)R_FLDAT(dev_num) = buf[i];
	}
	*R_FLMOD(dev_num) = org_mod;
	sflash_wait_busy(dev_num);
	CRITICAL_EXIT();
}

U32 RAMCODE sflash_custom_read(int dev_num, U8 cmd, U32 addr, U32 addr_byte_len,
							   int dummy_cnt, U8 *buf, U32 readlen)
{
	CRITICAL_ENTER();
	U32 i;
	U32 org_mod				 = *R_FLMOD(dev_num);
	*R_FLMOD(dev_num)		 = 0; // CS LOW, write-enable
	*(vU8 *)R_FLDAT(dev_num) = cmd;
	for (i = 0; i < addr_byte_len; i++)
		*(vU8 *)R_FLDAT(dev_num) =
			(addr >> (8 * (addr_byte_len - 1 - i))) & 0xff;

	for (i = 0; i < dummy_cnt; i++)
		*(vU8 *)R_FLDAT(dev_num) = 0; // dummy byte

	for (i = 0; i < readlen; i++)
		buf[i] = *(vU8 *)R_FLDAT(dev_num);

	*R_FLMOD(dev_num) = org_mod;
	CRITICAL_EXIT();
	return readlen;
}

void sflash_info_dump(int dev_num)
{
	debugprintf("============== FLASH[%d] information ==============\r\n",
				dev_num);
	debugprintf("    flash ids = 0x%x, 0x%x, 0x%x \r\n",
				flash_info[dev_num].ids[0], flash_info[dev_num].ids[1],
				flash_info[dev_num].ids[2]);
	debugprintf("    flash sector size = 0x%x \r\n",
				flash_info[dev_num].sector_size);
	debugprintf("    flash size = 0x%x \r\n", flash_info[dev_num].size);
	debugprintf("    SF Ctrl Register(0x%x) = 0x%x \r\n", R_FLMOD(dev_num),
				*R_FLMOD(dev_num));
	debugprintf("    SF Ctrl Register(0x%x) = 0x%x \r\n", R_FLBRT(dev_num),
				*R_FLBRT(dev_num));
	debugprintf("    SF Ctrl Register(0x%x) = 0x%x \r\n", R_FLCSH(dev_num),
				*R_FLCSH(dev_num));
	debugstring("==================================================\r\n");
}

void sflash_protect_write(int dev_num, BOOL b_enable_protect)
{
	if (b_enable_protect)
		WRITE_PROTECT_ENABLE(dev_num);
	else
		WRITE_PROTECT_DISABLE(dev_num);
}
